@@ -1,8 +1,18 @@
+#include "config.h"
+
+#include "../src/ss7_as.h"
+#include "../src/ss7_asp.h"
+#include "../src/ss7_link.h"
+#include "../src/ss7_linkset.h"
+#include "../src/ss7_route.h"
+#include "../src/ss7_route_table.h"
+#include "../src/ss7_user.h"
 #include "../src/xua_internal.h"
 #include "../src/xua_asp_fsm.h"
 
 #include <osmocom/sigtran/osmo_ss7.h>
 #include <osmocom/sigtran/protocol/m3ua.h>
+#include <osmocom/sigtran/protocol/sua.h>
 
 #include <osmocom/core/utils.h>
 #include <osmocom/core/msgb.h>
@@ -94,7 +104,7 @@ static int test_user_prim_cb(struct osmo_prim_hdr *oph, void *priv)
 
 static void test_user(void)
 {
-	struct osmo_ss7_user user, user2;
+	struct osmo_ss7_user *user, *user2;
 	struct osmo_mtp_prim omp = {
 		.oph = {
 			.sap = MTP_SAP_USER,
@@ -107,96 +117,120 @@ static void test_user(void)
 	};
 
 	printf("Testing SS7 user\n");
+	user = osmo_ss7_user_create(s7i, "testuser");
+	osmo_ss7_user_set_prim_cb(user, test_user_prim_cb);
+	osmo_ss7_user_set_priv(user, (void *) 0x1234);
 
-	user.name = "testuser";
-	user.priv = (void *) 0x1234;
-	user.prim_cb = test_user_prim_cb;
+	user2 = osmo_ss7_user_create(s7i, "testuser_notregistered");
+	osmo_ss7_user_set_prim_cb(user, test_user_prim_cb);
+	osmo_ss7_user_set_priv(user, (void *) 0x1234);
 
 	/* registration */
-	OSMO_ASSERT(osmo_ss7_user_register(s7i, 1, &user) == 0);
-	OSMO_ASSERT(osmo_ss7_user_register(s7i, 1, NULL) == -EBUSY);
-	OSMO_ASSERT(osmo_ss7_user_register(s7i, 255, NULL) == -EINVAL);
+	OSMO_ASSERT(osmo_ss7_user_register(user, 1) == 0);
+	OSMO_ASSERT(osmo_ss7_user_register(user, 1) == -EBUSY);
+	OSMO_ASSERT(osmo_ss7_user_register(user, 255) == -EINVAL);
 
 	/* primitive delivery */
-	OSMO_ASSERT(osmo_ss7_mtp_to_user(s7i, &omp) == 23);
+	OSMO_ASSERT(ss7_user_mtp_sap_prim_up(user, &omp) == 23);
 
 	/* cleanup */
-	OSMO_ASSERT(osmo_ss7_user_unregister(s7i, 255, NULL) == -EINVAL);
-	OSMO_ASSERT(osmo_ss7_user_unregister(s7i, 10, NULL) == -ENODEV);
-	OSMO_ASSERT(osmo_ss7_user_unregister(s7i, 1, &user2) == -EINVAL);
-	OSMO_ASSERT(osmo_ss7_user_unregister(s7i, 1, &user) == 0);
+	OSMO_ASSERT(osmo_ss7_user_unregister(user, 255) == -EINVAL);
+	OSMO_ASSERT(osmo_ss7_user_unregister(user, 10) == -ENODEV);
+	OSMO_ASSERT(osmo_ss7_user_unregister(user2, 1) == -EINVAL);
+	OSMO_ASSERT(osmo_ss7_user_unregister(user, 1) == 0);
 
-	/* primitive delivery should fail now */
-	OSMO_ASSERT(osmo_ss7_mtp_to_user(s7i, &omp) == -ENODEV);
-
-	/* wrong primitive delivery should also fail */
-	omp.oph.primitive = OSMO_MTP_PRIM_PAUSE;
-	OSMO_ASSERT(osmo_ss7_mtp_to_user(s7i, &omp) == -EINVAL);
+	osmo_ss7_user_destroy(user);
+	osmo_ss7_user_destroy(user2);
 }
 
+#define RT_LABEL(opc_val, dpc_val, sls_val) \
+	(struct osmo_ss7_route_label){ .opc = (opc_val), .dpc = (dpc_val), .sls = (sls_val) }
+
 static void test_route(void)
 {
 	struct osmo_ss7_route_table *rtbl;
 	struct osmo_ss7_linkset *lset_a, *lset_b;
+	struct osmo_ss7_link *l_a, *l_b;
 	struct osmo_ss7_route *rt, *rt12, *rtdef;
+	struct osmo_ss7_route_label route_label;
 
 	printf("Testing SS7 routing\n");
 
 	/* creation / destruction */
-	OSMO_ASSERT(osmo_ss7_route_table_find(s7i, "foobar") == NULL);
-	rtbl = osmo_ss7_route_table_find_or_create(s7i, "foobar");
+	OSMO_ASSERT(ss7_route_table_find(s7i, "foobar") == NULL);
+	rtbl = ss7_route_table_find_or_create(s7i, "foobar");
 	OSMO_ASSERT(rtbl);
-	OSMO_ASSERT(osmo_ss7_route_table_find_or_create(s7i, "foobar") == rtbl);
-	osmo_ss7_route_table_destroy(rtbl);
-	OSMO_ASSERT(osmo_ss7_route_table_find(s7i, "foobar") == NULL);
+	OSMO_ASSERT(ss7_route_table_find_or_create(s7i, "foobar") == rtbl);
+	ss7_route_table_destroy(rtbl);
+	OSMO_ASSERT(ss7_route_table_find(s7i, "foobar") == NULL);
 
 	/* we now work with system route table */
-	rtbl = osmo_ss7_route_table_find(s7i, "system");
+	rtbl = ss7_route_table_find(s7i, "system");
 	OSMO_ASSERT(rtbl && rtbl == s7i->rtable_system);
 
-	lset_a = osmo_ss7_linkset_find_or_create(s7i, "a", 100);
+	lset_a = ss7_linkset_find_or_create(s7i, "a", 100);
 	OSMO_ASSERT(lset_a);
-	lset_b = osmo_ss7_linkset_find_or_create(s7i, "b", 200);
+	l_a = ss7_link_find_or_create(lset_a, 0);
+	OSMO_ASSERT(l_a);
+	l_a->cfg.adm_state = OSMO_SS7_LS_ENABLED;
+
+	lset_b = ss7_linkset_find_or_create(s7i, "b", 200);
 	OSMO_ASSERT(lset_b);
+	l_b = ss7_link_find_or_create(lset_b, 0);
+	OSMO_ASSERT(l_b);
+	l_b->cfg.adm_state = OSMO_SS7_LS_ENABLED;
 
 	/* route with full mask */
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 12) == NULL);
-	rt = osmo_ss7_route_create(rtbl, 12, 0xffff, "a");
+	route_label = RT_LABEL(0, 12, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == NULL);
+	rt = ss7_route_create(rtbl, 12, 0xffff, false, "a");
 	printf("route with full mask: %s\n", osmo_ss7_route_print(rt));
 	OSMO_ASSERT(rt);
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 12) == rt);
-	osmo_ss7_route_destroy(rt);
+	route_label = RT_LABEL(0, 12, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == rt);
+	ss7_route_destroy(rt);
 
 	/* route with partial mask */
-	rt = osmo_ss7_route_create(rtbl, 8, 0xfff8, "a");
+	rt = ss7_route_create(rtbl, 8, 0xfff8, false, "a");
 	printf("route with partial mask: %s\n", osmo_ss7_route_print(rt));
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 8) == rt);
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 9) == rt);
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 12) == rt);
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 15) == rt);
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 16) == NULL);
+	route_label = RT_LABEL(0, 8, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == rt);
+	route_label = RT_LABEL(0, 9, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == rt);
+	route_label = RT_LABEL(0, 12, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == rt);
+	route_label = RT_LABEL(0, 15, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == rt);
+	route_label = RT_LABEL(0, 16, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == NULL);
 	/* insert more specific route for 12, must have higher priority
 	 * than existing one */
-	rt12 = osmo_ss7_route_create(rtbl, 12, 0xffff, "b");
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 12) == rt12);
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 15) == rt);
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 16) == NULL);
+	rt12 = ss7_route_create(rtbl, 12, 0xffff, false, "b");
+	route_label = RT_LABEL(0, 12, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == rt12);
+	route_label = RT_LABEL(0, 15, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == rt);
+	route_label = RT_LABEL(0, 16, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == NULL);
 	/* add a default route, which should have lowest precedence */
-	rtdef = osmo_ss7_route_create(rtbl, 0, 0, "a");
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 12) == rt12);
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 15) == rt);
-	OSMO_ASSERT(osmo_ss7_route_find_dpc(rtbl, 16) == rtdef);
+	rtdef = ss7_route_create(rtbl, 0, 0, false, "a");
+	route_label = RT_LABEL(0, 12, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == rt12);
+	route_label = RT_LABEL(0, 15, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == rt);
+	route_label = RT_LABEL(0, 16, 0);
+	OSMO_ASSERT(ss7_route_table_lookup_route(rtbl, &route_label) == rtdef);
 
-	osmo_ss7_route_destroy(rtdef);
-	osmo_ss7_route_destroy(rt12);
-	osmo_ss7_route_destroy(rt);
+	ss7_route_destroy(rtdef);
+	ss7_route_destroy(rt12);
+	ss7_route_destroy(rt);
 
-	rt = osmo_ss7_route_create(rtbl, 8, 0xfff9, "a");
+	rt = ss7_route_create(rtbl, 8, 0xfff9, false, "a");
 	printf("route with non-consecutive mask: %s\n", osmo_ss7_route_print(rt));
-	osmo_ss7_route_destroy(rt);
+	ss7_route_destroy(rt);
 
-	osmo_ss7_linkset_destroy(lset_a);
-	osmo_ss7_linkset_destroy(lset_b);
+	ss7_linkset_destroy(lset_a);
+	ss7_linkset_destroy(lset_b);
 }
 
 static void test_linkset(void)
@@ -206,38 +240,40 @@ static void test_linkset(void)
 
 	printf("Testing SS7 linkset/link\n");
 
-	OSMO_ASSERT(osmo_ss7_linkset_find_by_name(s7i, "a") == NULL);
-	OSMO_ASSERT(osmo_ss7_linkset_find_by_name(s7i, "b") == NULL);
+	OSMO_ASSERT(ss7_linkset_find_by_name(s7i, "a") == NULL);
+	OSMO_ASSERT(ss7_linkset_find_by_name(s7i, "b") == NULL);
 
-	lset_a = osmo_ss7_linkset_find_or_create(s7i, "a", 100);
+	lset_a = ss7_linkset_find_or_create(s7i, "a", 100);
 	OSMO_ASSERT(lset_a);
-	OSMO_ASSERT(osmo_ss7_linkset_find_by_name(s7i, "a") == lset_a);
+	OSMO_ASSERT(ss7_linkset_find_by_name(s7i, "a") == lset_a);
 
-	lset_b = osmo_ss7_linkset_find_or_create(s7i, "b", 200);
+	lset_b = ss7_linkset_find_or_create(s7i, "b", 200);
 	OSMO_ASSERT(lset_b);
-	OSMO_ASSERT(osmo_ss7_linkset_find_by_name(s7i, "b") == lset_b);
+	OSMO_ASSERT(ss7_linkset_find_by_name(s7i, "b") == lset_b);
 
-	l_a1 = osmo_ss7_link_find_or_create(lset_a, 1);
+	l_a1 = ss7_link_find_or_create(lset_a, 1);
 	OSMO_ASSERT(l_a1);
-	l_a2 = osmo_ss7_link_find_or_create(lset_a, 2);
+	l_a2 = ss7_link_find_or_create(lset_a, 2);
 	OSMO_ASSERT(l_a2);
 
 	/* ID too high */
-	OSMO_ASSERT(osmo_ss7_link_find_or_create(lset_a, 1000) == NULL);
+	OSMO_ASSERT(ss7_link_find_or_create(lset_a, 1000) == NULL);
 	/* already exists */
-	OSMO_ASSERT(osmo_ss7_link_find_or_create(lset_a, 1) == l_a1);
+	OSMO_ASSERT(ss7_link_find_or_create(lset_a, 1) == l_a1);
 
-	osmo_ss7_link_destroy(l_a1);
-	osmo_ss7_link_destroy(l_a2);
+	ss7_link_destroy(l_a1);
+	ss7_link_destroy(l_a2);
 
-	osmo_ss7_linkset_destroy(lset_a);
-	osmo_ss7_linkset_destroy(lset_b);
+	ss7_linkset_destroy(lset_a);
+	ss7_linkset_destroy(lset_b);
 }
 
 static void test_as(void)
 {
 	struct osmo_ss7_as *as;
 	struct osmo_ss7_asp *asp;
+	struct xua_msg *xua;
+	int rc;
 
 	OSMO_ASSERT(osmo_ss7_as_find_by_name(s7i, "as1") == NULL);
 	as = osmo_ss7_as_find_or_create(s7i, "as1", OSMO_SS7_ASP_PROT_M3UA);
@@ -258,9 +294,17 @@ static void test_as(void)
 	osmo_ss7_asp_restart(asp);
 
 	/* ask FSM to send ASP-UP.req */
-	osmo_fsm_inst_dispatch(asp->fi, XUA_ASP_E_M_ASP_UP_REQ, NULL);
-	osmo_fsm_inst_dispatch(asp->fi, XUA_ASP_E_ASPSM_ASPUP_ACK, NULL);
-	osmo_fsm_inst_dispatch(asp->fi, XUA_ASP_E_ASPTM_ASPAC_ACK, NULL);
+	rc = osmo_fsm_inst_dispatch(asp->fi, XUA_ASP_E_M_ASP_UP_REQ, NULL);
+	OSMO_ASSERT(rc == 0);
+
+	xua = xua_msg_alloc();
+	xua->hdr = XUA_HDR(SUA_MSGC_ASPSM, SUA_ASPSM_UP_ACK);
+	rc = osmo_fsm_inst_dispatch(asp->fi, XUA_ASP_E_ASPSM_ASPUP_ACK, xua);
+	xua_msg_free(xua);
+	OSMO_ASSERT(rc == 0);
+
+	rc = osmo_fsm_inst_dispatch(asp->fi, XUA_ASP_E_ASPTM_ASPAC_ACK, NULL);
+	OSMO_ASSERT(rc == 0);
 
 	OSMO_ASSERT(osmo_ss7_as_del_asp(as, "asp1") == 0);
 	OSMO_ASSERT(osmo_ss7_as_del_asp(as, "asp2") == -ENODEV);
